-------------------------------------------------------------------------------
--|> Lists
-- Manages the lists of players shown
-------------------------------------------------------------------------------
--|> UPVALUE GLOBALS
-------------------------------------------------------------------------------
local weizPVP = weizPVP
local InCombatLockdown, GetTime, RAID_CLASS_COLORS = InCombatLockdown, GetTime, RAID_CLASS_COLORS
local pairs, strlen, gsub, tinsert, sort = pairs, strlen, gsub, tinsert, sort
local C_Timer_After = C_Timer.After
local CopyTable = CopyTable

-------------------------------------------------------------------------------
--|> GLOBALS
-------------------------------------------------------------------------------
weizPVP.NearbyListSize = 0
weizPVP.CurrentList = {}
weizPVP.CurrentNameplates = {}
weizPVP.NearbyList = {}
weizPVP.ActiveList = {}
weizPVP.InactiveList = {}
weizPVP.ActiveDeadList = {}
weizPVP.InactiveDeadList = {}
weizPVP.PlayersOnBars = {}
-------------------------------------------------------------------------------
--|> LOCALS
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--|> FUNCTIONS
-------------------------------------------------------------------------------

--> Update Neaby Count
-------------------------------------------------------------------------------
local function UpdateNearbyCount()
	--> Upate Global
	weizPVP.NearbyCount = weizPVP.NearbyListSize

	if weizPVP.NearbyCount > 0 then --> List Size is > 0
		if weizPVP.NearbyCount > 50 then
			weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff0000" .. weizPVP.NearbyCount .. "|r") --> 50+ (RED)
		elseif weizPVP.NearbyCount > 40 then
			weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff0056" .. weizPVP.NearbyCount .. "|r") --> 40+ (BRIGHT PINK)
		elseif weizPVP.NearbyCount > 20 then
			weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff5900" .. weizPVP.NearbyCount .. "|r") --> 20+ (ORANGE)
		elseif weizPVP.NearbyCount > 10 then
			weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFFFBC00" .. weizPVP.NearbyCount .. "|r") --> 10+ (GOLD)
		else
			weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFd1ff00" .. weizPVP.NearbyCount .. "|r") --> 1+ (LIME GREEM)
		end
	else --> List Size is 0
		weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFF26FF540|r") --> 0 (GREEN)
	end
end

--> Manage Bars Displayed
-------------------------------------------------------------------------------
local listSize = 0
function weizPVP.ManageBarsDisplayed()
	listSize = weizPVP.NearbyListSize
	for i = 1, weizPVP.Options.Bars.MaxNumBars do
		if i > listSize then
			weizPVP.Bars[i]:SetAlpha(0)
			weizPVP.Bars[i]:SetValue(100)
			weizPVP.Bars[i].DEADIcon:Hide()
			weizPVP.Bars[i].RoleIcon:SetTexture("Interface\\AddOns\\weizPVP\\Icons\\unknown.tga", false)
			weizPVP.Bars[i].NameText:SetText("|cFFFFFFFF--|r")
		end
	end
end

--> Manage List Timeouts
-------------------------------------------------------------------------------
local expired = false
local removed = false
local expiredCount = 0
local i = 0
local timestamp = GetTime()
function weizPVP:ManageListTimeouts()
	--> Checks Active and Inactive lists for inactivity and move them accordingly
	--> Refreshed the list if we found expirations and Updates the NearbyCount

	expired = false
	removed = false
	expiredCount = 0
	i = 0
	timestamp = GetTime()

	for player in pairs(weizPVP.ActiveList) do
		if (timestamp - weizPVP.ActiveList[player].TimeUpdated) > weizPVP.Options.Sorting.NearbyActiveTimeout and weizPVP.CurrentNameplates[player] == nil then
			weizPVP.InactiveList[player] = weizPVP.ActiveList[player]
			weizPVP.InactiveList[player].TimeAdded = timestamp + (i * 0.001)
			weizPVP.ActiveList[player] = nil
			expired = true
			i = i + 1
		end
	end
	i = 0
	for player in pairs(weizPVP.ActiveDeadList) do
		if (timestamp - weizPVP.ActiveDeadList[player].TimeUpdated) > weizPVP.Options.Sorting.NearbyActiveTimeout then
			weizPVP.InactiveDeadList[player] = weizPVP.ActiveDeadList[player]
			weizPVP.InactiveDeadList[player].TimeAdded = timestamp + (i * 0.001)
			weizPVP.ActiveDeadList[player] = nil
			expired = true
			i = i + 1
		end
	end
	for player in pairs(weizPVP.InactiveList) do
		if (timestamp - weizPVP.InactiveList[player].TimeUpdated) > weizPVP.Options.Sorting.NearbyInactiveTimeout then
			weizPVP.InactiveList[player] = nil
			weizPVP.NearbyList[player] = nil
			weizPVP.PlayerActiveCache[player] = nil
			weizPVP.CurrentList[player] = nil
			expiredCount = expiredCount + 1
			expired = true
			removed = true
		end
	end

	for player in pairs(weizPVP.InactiveDeadList) do
		if (timestamp - weizPVP.InactiveDeadList[player].TimeUpdated) > weizPVP.Options.Sorting.NearbyInactiveTimeout then
			weizPVP.InactiveDeadList[player] = nil
			weizPVP.NearbyList[player] = nil
			weizPVP.PlayerActiveCache[player] = nil
			weizPVP.CurrentList[player] = nil
			expiredCount = expiredCount + 1
			expired = true
			removed = true
		end
	end

	if expired or removed then
		weizPVP.NearbyListSize = weizPVP.NearbyListSize - expiredCount
		UpdateNearbyCount()
		weizPVP:SortNearbyList()
		weizPVP:RefreshCurrentList()
	end

	if removed then
		weizPVP.ManageBarsDisplayed()
	end
end

--> Clear List Data
-------------------------------------------------------------------------------
function weizPVP:ClearListData()
	wipe(weizPVP.CurrentList)
	wipe(weizPVP.CurrentNameplates)
	wipe(weizPVP.NearbyList)
	wipe(weizPVP.ActiveList)
	wipe(weizPVP.InactiveList)
	wipe(weizPVP.ActiveDeadList)
	wipe(weizPVP.InactiveDeadList)
	wipe(weizPVP.PlayersOnBars)
	wipe(weizPVP.PlayerActiveCache)
	weizPVP.NearbyListSize = 0
	weizPVP:SortNearbyList()
	weizPVP:RefreshCurrentList()
	weizPVP.ManageBarsDisplayed()
	UpdateNearbyCount()
	weizPVP.TargetedIcon:SetAlpha(0)
	--> Refresh Target
	weizPVP:PlayerTargetEvent()
end

--> Format Level String
-------------------------------------------------------------------------------
local LevelText
local playerLevel
local function FormatLevelString(Estimated, Level)
	LevelText = ""
	playerLevel = weizPVP.Player.Level
	if Level then
		--> Known Level: Get colors based on difference
		if Level == 0 then --> 0
			LevelText = "|cFFFF00CC??|r"
		elseif Level < playerLevel - 20 then --> 20+ below
			LevelText = "|cFF7cffd1" .. Level .. "|r"
		elseif Level < playerLevel - 10 then --> 10-20 below
			LevelText = "|cFF7cd1ff" .. Level .. "|r"
		elseif Level < playerLevel then --> 1-9 below
			LevelText = "|cFF7cff7f" .. Level .. "|r"
		elseif Level > playerLevel then --> Higher level
			LevelText = "|cFFf7694a" .. Level .. "|r"
		else
			LevelText = "|cFFffc863" .. Level .. "|r"
		end
	end

	if Estimated == true and Level ~= 0 then
		LevelText = LevelText .. "|cFFFF00CC+|r"
	end

	return LevelText
end

--> Update Bar
local Bar
local Button
local darkenValue = -0.05
local charname
local classColor
local adjustedRole
local guildTxtLength
-------------------------------------------------------------------------------
--> Updates the player bar for the bar# specified
local function UpdateBar(num, player, Alpha, Health, Class, Guild, Level, Estimated, Stealth, Dead, Role, GUID, Name)
	Bar = weizPVP.Bars[num]
	Button = weizPVP.Buttons[num]
	if Button ~= nil and Bar ~= nil and GUID ~= nil then
		if not InCombatLockdown() and player and Name then --!! need to find the root of the issue
			Button:SetAttribute("macrotext", "/target " .. Name)
			Bar.Target = Name
		end

		Button:SetAlpha(0.2)
		if Alpha and Alpha ~= 0 then
			Bar:SetAlpha(Alpha)
		else
			Bar:SetAlpha(weizPVP.Options.Bars.AlphaDefault)
		end

		--> NAME TEXT
		charname = gsub(Name, "-(.*)", "")
		if strlen(Name) > strlen(charname) then
			Bar.NameText:SetText(charname .. "|cFFFF00CC*|r")
		else
			Bar.NameText:SetText(Name)
		end
		Bar.name = Name
		Bar.GUID = GUID
		weizPVP.PlayersOnBars[GUID] = num

		--> Stealth
		if Stealth ~= nil and Stealth == true then
			if strlen(Name) > strlen(charname) then
				Bar.NameText:SetText("|cFFffc863" .. charname .. "|r |cFFFF00CC*|r")
			else
				Bar.NameText:SetText("|cFFffc863" .. Name .. "|r")
			end
		end
		--> LEVEL
		Bar.LevelText:SetText(FormatLevelString(Estimated, Level))

		--> CLASS + COLOR (BAR COLOR)
		if Class and RAID_CLASS_COLORS[Class] then
			classColor = RAID_CLASS_COLORS[Class]
			Bar:SetStatusBarColor(classColor.r - darkenValue, classColor.g - darkenValue, classColor.b - darkenValue, Alpha)
		end
		Bar.bg:SetVertexColor(0, 0, 0, 1)

		--> HEALTH (BAR VALUE)
		if Health ~= nil then
			Bar:SetValue(Health)
			if Health > 0 then
				Dead = false
			else
				Dead = true
			end
		end

		--> ROLE
		adjustedRole = weizPVP:RoleAssign(Class) or Role
		if adjustedRole == "HEALER" then
			Bar.RoleIcon:SetTexture("Interface\\AddOns\\weizPVP\\Icons\\healer.tga", false)
		elseif adjustedRole == "DAMAGER" then
			Bar.RoleIcon:SetTexture("Interface\\AddOns\\weizPVP\\Icons\\damager.tga", false)
		elseif adjustedRole == "TANK" then
			Bar.RoleIcon:SetTexture("Interface\\AddOns\\weizPVP\\Icons\\tank.tga", false)
		else
			Bar.RoleIcon:SetTexture("Interface\\AddOns\\weizPVP\\Icons\\unknown.tga", false)
		end

		--> DEAD ICON
		if Dead ~= nil then
			if Dead == true then
				Bar.DEADIcon:Show()
				Bar:SetValue(0)
			elseif Dead == false then
				Bar.DEADIcon:Hide()
			end
		end

		--> GUILD TEXT
		-----------------------------------------------------------------
		guildTxtLength = Bar.LevelText:GetWidth() + Bar.NameText:GetWidth() + 28
		if Bar.DEADIcon:IsShown() then
			guildTxtLength = guildTxtLength + 22
		end

		guildTxtLength = Bar:GetWidth() - guildTxtLength
		Bar.GuildText:SetText(Guild)
		Bar.GuildText:SetWidth(guildTxtLength)
	else
		Bar:SetAlpha(weizPVP.Options.Bars.AlphaDefault)
	end
end

--> Refresh One Bar
-------------------------------------------------------------------------------
local row
local Alpha
local function RefreshOneBar(player)
	--> Get Row
	row = weizPVP.PlayersOnBars[player]

	--> Alpha
	Alpha = weizPVP.Options.Bars.AlphaDefault or 0.8
	if weizPVP.InactiveList[player] or weizPVP.InactiveDeadList[player] then
		Alpha = weizPVP.Options.Bars.AlphaInactive
	end
	if weizPVP.ActiveDeadList[player] then
		Alpha = weizPVP.Options.Bars.AlphaDead
	end

	--> Update Bar
	if row then
		UpdateBar(
			row,
			player,
			Alpha,
			weizPVP.PlayerActiveCache[player].Health,
			weizPVP.PlayerActiveCache[player].Class,
			weizPVP.PlayerActiveCache[player].Guild,
			weizPVP.PlayerActiveCache[player].Level,
			weizPVP.PlayerActiveCache[player].Estimated,
			weizPVP.PlayerActiveCache[player].Stealth,
			weizPVP.PlayerActiveCache[player].Dead,
			weizPVP.PlayerActiveCache[player].Role,
			weizPVP.PlayerActiveCache[player].GUID,
			weizPVP.PlayerActiveCache[player].Name
		)
	end
end

--> Update Player List
-------------------------------------------------------------------------------
local reSortList
local playerOnBar
local function UpdatePlayerLists(player, timeUpdate, dead, newPlayerOnList)
	reSortList = false
	if weizPVP.NearbyListSize <= weizPVP.Options.Bars.MaxNumBars then
		if dead == true or newPlayerOnList == true then
			reSortList = true
		end
	else
		if weizPVP.PlayersOnBars[player] ~= nil then
			if dead == true then
				reSortList = true
			end
		end
	end

	playerOnBar = true

	if weizPVP.NearbyList[player] == nil then --> ADDING NEW PLAYER
		playerOnBar = false
		--> New Player added; play sound?
		if weizPVP.Options.AudioAlerts.DetectedPlayerSound == true then
			weizPVP.PlaySoundAlert("Interface\\AddOns\\weizPVP\\Sounds\\detected-nearby.ogg")
		end

		weizPVP.NearbyList[player] = {}
		weizPVP.NearbyList[player].TimeUpdated = timeUpdate
		weizPVP.NearbyList[player].TimeAdded = timeUpdate

		if dead == true then
			weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
			weizPVP.ActiveDeadList[player].TimeUpdated = timeUpdate
			weizPVP.ActiveDeadList[player].TimeAdded = timeUpdate
		else
			weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
			weizPVP.ActiveList[player].TimeAdded = timeUpdate
			weizPVP.ActiveList[player].TimeUpdated = timeUpdate
		end
	elseif not weizPVP.ActiveList[player] and not weizPVP.ActiveDeadList[player] then --> EXISTING PLAYER; WAS INACTIVE
		reSortList = true
		if dead then
			weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
			weizPVP.ActiveDeadList[player].TimeUpdated = timeUpdate
			weizPVP.ActiveDeadList[player].TimeAdded = timeUpdate
			weizPVP.NearbyList[player].TimeUpdated = timeUpdate
			weizPVP.NearbyList[player].TimeAdded = timeUpdate
			weizPVP.ActiveList[player] = nil
		else
			weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
			weizPVP.ActiveList[player].TimeUpdated = timeUpdate
			weizPVP.ActiveList[player].TimeAdded = timeUpdate
			weizPVP.NearbyList[player].TimeUpdated = timeUpdate
			weizPVP.NearbyList[player].TimeAdded = timeUpdate
			weizPVP.ActiveDeadList[player] = nil
		end

		weizPVP.InactiveList[player] = nil
		weizPVP.InactiveDeadList[player] = nil
	else --> EXISTING PLAYER; ACTIVE
		if dead then
			weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
			weizPVP.ActiveDeadList[player].TimeUpdated = timeUpdate
			weizPVP.NearbyList[player].TimeUpdated = timeUpdate
			weizPVP.ActiveList[player] = nil
		else
			weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
			weizPVP.ActiveList[player].TimeUpdated = timeUpdate
			weizPVP.NearbyList[player].TimeUpdated = timeUpdate
			weizPVP.ActiveDeadList[player] = nil
		end
		weizPVP.InactiveList[player] = nil
		weizPVP.InactiveDeadList[player] = nil
	end

	if playerOnBar == false and weizPVP.NearbyListSize > weizPVP.Options.Bars.MaxNumBars then
		return
	end

	--> Sort only if we moved the player from one sublist to another
	if reSortList == true then
		weizPVP:SortNearbyList()
		if weizPVP.PlayersOnBars[player] or newPlayerOnList then
			weizPVP:RefreshCurrentList()
		end
	else
		if weizPVP.PlayersOnBars[player] then
			RefreshOneBar(player)
		end
	end
end

--> Add Player Data
--> Incoming player data from event
-------------------------------------------------------------------------------
local lastTimestamp = GetTime()
local dead
local currentTime = GetTime()
function weizPVP:AddPlayerDataToNearby(GUID, newPlayerOnList)
	if GUID and weizPVP.PlayerActiveCache[GUID] then
		if weizPVP.PlayerActiveCache[GUID].Dead then
			dead = weizPVP.PlayerActiveCache[GUID].Dead
		else
			weizPVP.PlayerActiveCache[GUID].Dead = false
			dead = false
		end
	else
		return
	end
	currentTime = GetTime()
	if lastTimestamp >= currentTime then
		timestamp = lastTimestamp + 0.001
		lastTimestamp = timestamp
	else
		lastTimestamp = currentTime
		timestamp = lastTimestamp
	end
	if newPlayerOnList then
		weizPVP.NearbyListSize = weizPVP.NearbyListSize + 1
		UpdateNearbyCount()
	end
	UpdatePlayerLists(GUID, timestamp, dead, newPlayerOnList)
end

--> Refresh Current List
-------------------------------------------------------------------------------
local playersOnBars
function weizPVP:RefreshCurrentList()
	playersOnBars = {}
	--> Refreshing all bars due to required sorting
	for k, data in pairs(weizPVP.CurrentList) do
		if k <= weizPVP.Options.Bars.MaxNumBars then
			--> Update Alpha if needed
			Alpha = weizPVP.Options.Bars.AlphaDefault
			if weizPVP.InactiveList[data.player] or weizPVP.InactiveDeadList[data.player] then
				Alpha = weizPVP.Options.Bars.AlphaInactive
			end
			if weizPVP.ActiveDeadList[data.player] then
				Alpha = weizPVP.Options.Bars.AlphaDead
			end
			playersOnBars[weizPVP.PlayerActiveCache[data.player].GUID] = k
			--> Update Bar
			UpdateBar(
				k,
				data.player,
				Alpha,
				weizPVP.PlayerActiveCache[data.player].Health,
				weizPVP.PlayerActiveCache[data.player].Class,
				weizPVP.PlayerActiveCache[data.player].Guild,
				weizPVP.PlayerActiveCache[data.player].Level,
				weizPVP.PlayerActiveCache[data.player].Estimated,
				weizPVP.PlayerActiveCache[data.player].Stealth,
				weizPVP.PlayerActiveCache[data.player].Dead,
				weizPVP.PlayerActiveCache[data.player].Role,
				weizPVP.PlayerActiveCache[data.player].GUID,
				weizPVP.PlayerActiveCache[data.player].Name
			)
		else
			if playersOnBars then
				wipe(weizPVP.PlayersOnBars)
				weizPVP.PlayersOnBars = CopyTable(playersOnBars)
			end
			break
		end
	end
	weizPVP:ChangeTargetIcon()
end

--> Sort Nearby List
-------------------------------------------------------------------------------
local position
local active = {}
local activeDead = {}
local inactiveDead = {}
local inactive = {}
function weizPVP:SortNearbyList()
	--=> Get player's into the right groups
	--> ActiveList
	active = {}
	for player in pairs(weizPVP.ActiveList) do
		if weizPVP.NearbyList[player] then
			position = weizPVP.NearbyList[player].TimeAdded or nil
			if position ~= nil then
				tinsert(active, {player = player, time = position})
			end
		end
	end

	--> ActiveDeadList
	activeDead = {}
	for player in pairs(weizPVP.ActiveDeadList) do
		if weizPVP.NearbyList[player] then
			position = weizPVP.NearbyList[player].TimeAdded or nil
			if position ~= nil then
				tinsert(activeDead, {player = player, time = position})
			end
		end
	end

	--> InactiveDeadList
	inactiveDead = {}
	for player in pairs(weizPVP.InactiveDeadList) do
		if weizPVP.NearbyList[player] then
			position = weizPVP.NearbyList[player].TimeAdded or nil
			if position ~= nil then
				tinsert(inactiveDead, {player = player, time = position})
			end
		end
	end

	--> InactiveList
	inactive = {}
	for player in pairs(weizPVP.InactiveList) do
		if weizPVP.NearbyList[player] then
			position = weizPVP.NearbyList[player].TimeAdded
			if position ~= nil then
				tinsert(inactive, {player = player, time = position})
			end
		end
	end

	sort(
		active,
		function(a, b)
			return a.time < b.time
		end
	)

	sort(
		activeDead,
		function(a, b)
			return a.time < b.time
		end
	)

	sort(
		inactive,
		function(a, b)
			return a.time < b.time
		end
	)

	sort(
		inactiveDead,
		function(a, b)
			return a.time < b.time
		end
	)

	local list = {}

	for player in pairs(active) do
		tinsert(list, active[player])
	end

	for player in pairs(activeDead) do
		tinsert(list, activeDead[player])
	end

	for player in pairs(inactive) do
		tinsert(list, inactive[player])
	end
	for player in pairs(inactiveDead) do
		tinsert(list, inactiveDead[player])
	end

	weizPVP.CurrentList = list
end

--> Play Sound Alert
-------------------------------------------------------------------------------
local playSound = true
local function EnableAlertSound()
	playSound = true
end
function weizPVP.PlaySoundAlert(url)
	if playSound == true and weizPVP.Options.AudioAlerts.DetectedPlayerSound == true then
		PlaySoundFile(url)
		playSound = false
		C_Timer_After(1, EnableAlertSound)
	end
end
